const { readdir, stat, createReadStream } = require('fs');
const { join, parse } = require('path');

/**
 * Takes a number of a file size in bytes and converts it to a human readable format
 * @param {number} size A number to be converted
 * @returns {string} A size string formatted
 */
const bytesToHumanReadable = size => {
  let i = size == 0 ? 0 : Math.floor(Math.log(size) / Math.log(1024));
  return (size / Math.pow(1024, i)).toFixed(2) * 1 + ' ' + ['Bytes', 'kB', 'MB', 'GB', 'TB'][i];
};

/**
 *
 * @param {string} directory The file path of a folder e.g. './photos/'
 */
const readDirectory = directory => {
  return new Promise((resolve, reject) => {
    readdir(directory, (err, files) => {
      if (err) return reject(err);
      return resolve({ files: files, directory: directory });
    });
  });
};

/**
 * adds file size, file path, if it's a folder, creation date and file extension to files resolved from readDirectory()
 * @param {[{files: [string], directory: string}]} files
 */
const addStatistics = files => {
  return new Promise((resolve, reject) => {
    files['files'].forEach((file, index) => {
      const filePath = join(files.directory, file);
      const fileName = parse(file).name;
      files['files'][index] = {
        name: fileName,
        path: filePath
      };
      stat(filePath, (err, stats) => {
        if (err) return reject(err);
        files['files'][index]['statistics'] = {
          size: stats.size,
          sizeHuman: bytesToHumanReadable(stats.size),
          directory: stats.isDirectory(),
          date: new Date(stats.birthtime),
          extension: parse(filePath).ext
        };
        if (index >= files['files'].length - 1) return resolve(files);
      });
    });
  });
};

/**
 * Reads a directory for files/folders with their respective statistics
 * @param {string} directory The directory to read files from
 * @param {boolean} traversalProtection Protection from malicious users trying to escape the file delivery folder
 * @param {function} callback A callback function that takes an error as first parameter and the result as a second object
 * @return {[{}]} The readings of the directory
 */
const readDir = (directory = './public', traversalProtection = true, callback) => {
  if ((traversalProtection && directory.includes('..')) || directory.toUpperCase().includes('%2E'))
    return callback({ msg: 'Permission Denied', status: 403, code: 'DIRECTORY_TRAVERSAL' }, null);
  readDirectory(directory)
    .then(addStatistics)
    .then(files => callback(null, files))
    .catch(err => callback(err, null));
};

/**
 * Sends (for a client to download) a file from a static directory
 * @param {string} directory The absolute path of the file
 * @param {boolean} traversalProtection Protection from malicious users trying to escape the file delivery folder
 * @param {function} callback A callback function that takes an error as first parameter and a readable stream as a second object
 */
const sendFile = (directory, traversalProtection = true, callback) => {
  if ((traversalProtection && directory.includes('..')) || directory.toUpperCase().includes('%2E'))
    return callback({ msg: 'Permission Denied', status: 403, code: 'DIRECTORY_TRAVERSAL' }, null);
  if (!directory) return callback({ msg: 'Path to a file must be specified', status: 500, code: 'PATH_NOT_FOUND' }, null);
  let readable;
  try {
    readable = createReadStream(directory);
  } catch (err) {
    return callback({ msg: 'File not found', status: 404, code: 'FILE_NOT_FOUND' }, null);
  }
  readable.on('data', chunk => {
    callback(null, chunk);
  });
};

module.exports = { readDir, sendFile };
